// -*- C++ -*-
/*!
 * @file  InPortProvider.h
 * @brief InPortProvider class
 * @date  $Date: 2007-12-31 03:08:03 $
 * @author Noriaki Ando <n-ando@aist.go.jp>
 *
 * Copyright (C) 2006-2010
 *     Noriaki Ando
 *     Task-intelligence Research Group,
 *     Intelligent Systems Research Institute,
 *     National Institute of
 *         Advanced Industrial Science and Technology (AIST), Japan
 *     All rights reserved.
 *
 * $Id: InPortProvider.h 1878 2010-02-17 08:02:20Z n-ando $
 *
 */

#ifndef RTC_INPORTPROVIDER_H
#define RTC_INPORTPROVIDER_H

#include <string>

#include <coil/Factory.h>

#include <rtm/BufferBase.h>
#include <rtm/NVUtil.h>
#include <rtm/SystemLogger.h>
#include <rtm/DataPortStatus.h>
#include <rtm/InPortConnector.h>

namespace RTC
{
  class ConnectorListeners;
  class ConnectorInfo;

  /*!
   * @if jp
   *
   * @class OutPortProvider
   * @brief OutPortProvider
   *
   * InPort  PROVIDED C^[tF[X邽߂̒ۊNXB
   * InPort ɑ΂ĐVC^[tF[Xꍇɂ́ÃN
   * XpAȉ̊֐KvB
   *
   * - init()
   * - setBuffer()
   * - setListener()
   * - setConnector()
   * 
   * ɁARXgN^ňȉ̊֐ĂсAݒK
   * vB
   * 
   * - setInterfaceType()
   * - setDataFlowType()
   * - setSubscriptionType()
   *
   * ̂قAInPortProvider ̃vpeBƂĊOɌJKv̂
   * ĺAprotected ϐ (SDOPackage::NVList) m_properties ɑ΂
   * Zbg邱ƁBZbgꂽĺAC^[tF[X̃vt@C
   * āA܂Aڑɑ̃C^[tF[Xɂ̃C^[tF[XɊ
   * ^ۂɗpBȉ̉z֐́A|[g̃C^[
   * tF[Xvt@C擾ѐڑɃ|[gĂяoB
   * \߃Zbgꂽ̃C^[tF[X̃vt@C͂̊
   * Ăяoɂ|[gɓ`B
   *
   * - publishInterfaceProfile()
   * - publishInterface()
   *
   * InPort  InPortProvider ̃t@NgǗNXɑ΂ėp\
   *  InPortProvider ⍇A񋟉\ȃC^[tF[X^CvO
   * ɐ錾B]āAInPort@ɑ΂ PROVIDED C^[tF[X
   * 񋟂 InPortProvider ̃TuNX́AInPortProviderFactory
   * Ƀt@Ng֐o^KvB
   *
   * RTC::InPortProviderFactory::instance().addFactory() A
   *
   * - 1: voC_̖O, "corba_cdr" Ȃ
   * - 2: t@Ng֐ coil::Creator<B, T>
   * - 3: 폜֐ coil::Destructor<B, T>
   * 
   * ^ČĂяoKvBȉ́At@Ngւ̓o^ƁA
   * ֐ƂłB
   * 
   * <pre>
   * extern "C"
   * {
   *   void InPortCorbaCdrProviderInit(void)
   *   {
   *     RTC::InPortProviderFactory&
   *                         factory(RTC::InPortProviderFactory::instance());
   *     factory.addFactory("corba_cdr",
   *                        ::coil::Creator<::RTC::InPortProvider,
   *                                        ::RTC::InPortCorbaCdrProvider>,
   *                        ::coil::Destructor<::RTC::InPortProvider,
   *                                           ::RTC::InPortCorbaCdrProvider>);
   *   }
   * };
   * </pre>
   *
   * ̗̂悤ɁAt@Ngւ̓o^֐ƂāAextern "C"
   * ɂV{QƉ\ɂĂB邱ƂŁA
   * InPortProvider LIuWFNg (DLL) ēI[h\
   * AvoC_̌^𓮓Iɒǉ邱Ƃ\ƂȂB
   *
   * @since 0.4.0
   *
   * @else
   *
   * @class InPortProvider
   * @brief InPortProvider
   *
   * The virtual class for InPort's PROVIDED interface
   * implementation.  New interface for InPort have to inherit this
   * class, and have to implement the following functions.
   *
   * - init()
   * - setBuffer()
   * - setListener()
   * - setConnector()
   * 
   * Moreover, calling the following functions in the constructor, and
   * properties have to be set.
   *
   * - setPortType()
   * - setDataType()
   * - setInterfaceType()
   * - setDataFlowType()
   * - setSubscriptionType()
   *
   * InPortProvider's properties that have to be provided to others
   * should be set to protected variable (SDOPackage::NVList)
   * m_properties. Values that are set to the property are published
   * as interface profile information, and it is also published to
   * required interface when connection is established.  The following
   * virtual functions are called when port's profiles are acquired
   * from others or connections are established. The following virtual
   * functions are called when port's profiles are acquired from
   * others or connections are established. Interface profile
   * information that is reviously set is given to Port calling by
   * these functions.
   *
   * - publishInterfaceProfile()
   * - publishInterface()
   *
   * InPort inquires available InPortProviders to the factory class
   * of InPortProvider, and publishes available interfaces to
   * others. Therefore, sub-classes of InPortProvider that provides
   * PROVIDED interface to InPort should register its factory to
   * InPortProviderFactory.
   *
   * RTC::InPortProviderFactory::instance().addFactory() would be
   * called with the following arguments.
   *
   * 1st arg: The name of provider. ex. "corba_cdr"
   * 2nd arg: Factory function. coil::Creator<B, T>
   * 3rd arg: Destruction function. coil::Destructor<B, T>
   *
   * The following example shows how to register factory function.
   * And it is also declared as a initialization function.
   *
   * <pre>
   * extern "C"
   * {
   *   void InPortCorbaCdrProviderInit(void)
   *   {
   *     RTC::InPortProviderFactory&
   *                         factory(RTC::InPortProviderFactory::instance());
   *     factory.addFactory("corba_cdr",
   *                        ::coil::Creator<::RTC::InPortProvider,
   *                                        ::RTC::InPortCorbaCdrProvider>,
   *                        ::coil::Destructor<::RTC::InPortProvider,
   *                                           ::RTC::InPortCorbaCdrProvider>);
   *   }
   * };
   * </pre>
   *
   * It is recommended that the registration process is declared as a
   * initialization function with "extern C" to be accessed from the
   * outside of module.  If the InPortProviders are compiled as a
   * shared object or DLL for dynamic loading, new InPortProvider
   * types can be added dynamically.
   *
   * @since 0.4.0
   *
   * @endif
   */
  class InPortProvider
    : public DataPortStatus
  {
  public:
    DATAPORTSTATUS_ENUM
    /*!
     * @if jp
     * @brief RXgN^
     *
     * RXgN^
     *
     * @else
     * @brief Constructor
     *
     * Constructor
     *
     * @endif
     */
    InPortProvider();
    
    /*!
     * @if jp
     * @brief fXgN^
     *
     * zfXgN^
     *
     * @else
     * @brief Destructor
     *
     * Virtual destructor
     *
     * @endif
     */
    virtual ~InPortProvider(void);

    /*!
     * @if jp
     * @brief ݒ菉
     *
     * OutPortProvider ̊eݒsBNXł́A^ꂽ
     * PropertiesKvȏ擾ĊeݒsB init() 
     * ́AOutPortProviderエсAڑɂꂼĂ΂
     * \BāÅ֐͕Ă΂邱Ƃz肵ċL
     * qׂłB
     * 
     * @param prop ݒ
     *
     * @else
     *
     * @brief Initializing configuration
     *
     * This operation would be called to configure in initialization.
     * In the concrete class, configuration should be performed
     * getting appropriate information from the given Properties data.
     * This function might be called right after instantiation and
     * connection sequence respectivly.  Therefore, this function
     * should be implemented assuming multiple call.
     *
     * @param prop Configuration information
     *
     * @endif
     */
    virtual void init(coil::Properties& prop) = 0;

    /*!
     * @if jp
     * @brief obt@Zbg
     *
     * OutPortProviderf[^oobt@ZbgB
     * łɃZbgꂽobt@ꍇAȑÕobt@ւ
     * |C^ɑ΂ď㏑B
     * OutPortProvider̓obt@̏L肵ĂȂ̂ŁA
     * obt@̍폜̓[U̐ӔCōsȂ΂ȂȂB
     *
     * @param buffer OutPortProviderf[^oobt@ւ̃|C^
     *
     * @else
     * @brief Setting outside buffer's pointer
     *
     * A pointer to a buffer from which OutPortProvider retrieve data.
     * If already buffer is set, previous buffer's pointer will be
     * overwritten by the given pointer to a buffer.  Since
     * OutPortProvider does not assume ownership of the buffer
     * pointer, destructor of the buffer should be done by user.
     * 
     * @param buffer A pointer to a data buffer to be used by OutPortProvider
     *
     * @endif
     */
    virtual void setBuffer(BufferBase<cdrMemoryStream>* buffer) = 0;

    /*!
     * @if jp
     * @brief Xiݒ肷B
     *
     * OutPort ̓f[^MɂeCxgɑ΂ē̃Xi
     * IuWFNgR[R[obN@\񋟂Bڍׂ
     * ConnectorListener.h  ConnectorDataListener, ConnectorListener
     * QƂ̂ƁBOutPortProvider ̃TuNXł́A^ꂽX
     * iK؂ȃ^C~OŌĂяoׂłBAׂẴXi
     * ĂяoKv͂ȂB
     *
     * @param info ڑ
     * @param listeners XiIuWFNg
     *
     * @else
     * @brief Set the listener. 
     *
     * OutPort provides callback functionality that calls specific
     * listener objects according to the events in the data publishing
     * process. For details, see documentation of
     * ConnectorDataListener class and ConnectorListener class in
     * ConnectorListener.h. In the sub-classes of OutPortProvider, the
     * given listeners should be called in the proper timing. However,
     * it is not necessary to call all the listeners.
     *
     * @param info Connector information
     * @param listeners Listener objects
     *
     * @endif
     */
    virtual void setListener(ConnectorInfo& info,
                             ConnectorListeners* listeners) = 0;

    /*!
     * @if jp
     * @brief Connectorݒ肷B
     *
     * OutPort ͐ڑm OutPortConnector IuWFNg𐶐A
     * IuWFNg̃|C^Ƌɂ̊֐ĂяoBL
     * OutPort ێ̂ OutPortProvider  OutPortConnector 
     * Ă͂ȂB
     *
     * @param connector OutPortConnector
     *
     * @else
     * @brief set Connector
     *
     * OutPort creates OutPortConnector object when it establishes
     * connection between OutPort and InPort, and it calls this
     * function with a pointer to the connector object. Since the
     * OutPort has the ownership of this connector, OutPortProvider
     * should not delete it.
     *
     * @param connector OutPortConnector
     *
     * @endif
     */
    virtual void setConnector(InPortConnector* connector) = 0;

    /*!
     * @if jp
     * @brief InterfaceProfileJ
     *
     * InterfaceProfileJB
     * Ŏw肷vpeB NameValue IuWFNg
     * dataport.interface_type l𒲂ׁAY|[gɐݒ肳Ă
     * C^[tF[X^Cvƈvꍇ̂ݏ擾B
     *
     * @param properties InterfaceProfile󂯎vpeB
     *
     * @else
     * @brief Publish InterfaceProfile information
     *
     * Publish interfaceProfile information.
     * Check the dataport.interface_type value of the NameValue object 
     * specified by an argument in property information and get information
     * only when the interface type of the specified port is matched.
     *
     * @param properties Properties to get InterfaceProfile information
     *
     * @endif
     */
    virtual void publishInterfaceProfile(SDOPackage::NVList& properties);
    
    /*!
     * @if jp
     * @brief InterfaceJ
     *
     * InterfaceJBŎw肷vpeB
     * NameValue IuWFNgdataport.interface_type l𒲂ׁAY|[
     * gɐݒ肳ĂȂNameValue ɏǉBłɓC
     * ^[tF[Xo^ς݂̏ꍇ͉sȂB
     *
     * @param properties Interface󂯎vpeB
     * @return true: I
     *
     * @else
     * @brief Publish interface information
     *
     * Publish interface information.  Check the
     * dataport.interface_type value of the NameValue object specified
     * by an argument in the property information, and add the
     * information to the NameValue if this port is not specified.
     * This does not do anything if the same interface is already
     * subscribed.
     *
     * @param properties Properties to receive interface information
     * @return true: normal return
     *
     * @endif
     */
    virtual bool publishInterface(SDOPackage::NVList& properties);
    
  protected:
    /*!
     * @if jp
     * @brief C^tF[X^Cvݒ肷
     *
     * C^tF[X^Cvݒ肷B
     *
     * @param interface_type ݒΏۃC^tF[X^Cv
     *
     * @else
     * @brief Set the interface type
     *
     * Set the interface type.
     *
     * @param interface_type The target interface type for set
     *
     * @endif
     */
    void setInterfaceType(const char* interface_type);
    
    /*!
     * @if jp
     * @brief f[^t[^Cvݒ肷
     *
     * f[^t[^Cvݒ肷B
     *
     * @param dataflow_type ݒΏۃf[^t[^Cv
     *
     * @else
     * @brief Set the dataflow type
     *
     * Set the dataflow type.
     *
     * @param dataflow_type The target data flow type for set
     *
     * @endif
     */
    void setDataFlowType(const char* dataflow_type);
    
    /*!
     * @if jp
     * @brief TuXNvV^Cvݒ肷
     *
     * TuXNvV^Cvݒ肷B
     *
     * @param subs_type ݒΏۃTuXNvV^Cv
     *
     * @else
     * @brief Set the subscription type
     *
     * Set the subscription type.
     *
     * @param subs_type The target subscription type for set
     *
     * @endif
     */
    void setSubscriptionType(const char* subs_type);


  protected:
    /*!
     * @if jp
     * @brief |[gvt@CێvpeB
     * @else
     * @brief Properties to hold port profile
     * @endif
     */
    SDOPackage::NVList m_properties;

    /*!
     * @if jp
     * @brief K[Xg[
     * @else
     * @brief Logger stream
     * @endif
     */
    mutable Logger rtclog;

  private:
    std::string m_interfaceType;
    std::string m_dataflowType;
    std::string m_subscriptionType;

  public:
    // functors
    /*!
     * @if jp
     * @brief C^[tF[Xvt@CJ邽̃t@N^
     * @else
     * @brief Functor to publish interface profile
     * @endif
     */
    struct publishInterfaceProfileFunc
    {
      publishInterfaceProfileFunc(SDOPackage::NVList& prop) : m_prop(prop) {}
      void operator()(InPortProvider* provider)
      {
	provider->publishInterfaceProfile(m_prop);
      }
      SDOPackage::NVList& m_prop;
    };

    /*!
     * @if jp
     * @brief C^[tF[Xvt@CJ邽̃t@N^
     * @else
     * @brief Functor to publish interface profile
     * @endif
     */
    struct publishInterfaceFunc
    {
      publishInterfaceFunc(SDOPackage::NVList& prop)
        : m_prop(prop), provider_(0) {}
      void operator()(InPortProvider* provider)
      {
	if (provider->publishInterface(m_prop))
          {
            provider_ = provider;
          }
      }
      SDOPackage::NVList& m_prop;
      InPortProvider* provider_;
    };


  };

  /*!
   * @if jp
   * @brief InPortProviderFactory^錾
   * @else
   * @brief InPortProviderFactory type definition
   * @endif
   */
  typedef ::coil::GlobalFactory<InPortProvider> InPortProviderFactory;

};     // namespace RTC
#endif // RTC_INPORTPROVIDER_H
