// -*- C++ -*-
/*!
 * @file RingBuffer.h
 * @brief Defautl Buffer class
 * @date $Date: 2007-12-31 03:08:06 $
 * @author Noriaki Ando <n-ando@aist.go.jp>
 *
 * Copyright (C) 2006-2009
 *     Noriaki Ando
 *     Task-intelligence Research Group,
 *     Intelligent Systems Research Institute,
 *     National Institute of
 *         Advanced Industrial Science and Technology (AIST), Japan
 *     All rights reserved.
 *
 * $Id: RingBuffer.h 1988 2010-07-08 03:17:01Z kurihara $
 *
 */

#ifndef RTC_RINGBUFFER_H
#define RTC_RINGBUFFER_H

#include <vector>
#include <algorithm>
#include <iostream>

#include <coil/TimeValue.h>
#include <coil/Mutex.h>
#include <coil/Guard.h>
#include <coil/Condition.h>
#include <coil/stringutil.h>

#include <rtm/BufferBase.h>
#include <rtm/BufferStatus.h>

#define RINGBUFFER_DEFAULT_LENGTH 8
/*!
 * @if jp
 * @namespace RTC
 *
 * @brief RTR|[lg
 *
 * @else
 *
 * @namespace RTC
 *
 * @brief RT-Component
 *
 * @endif
 */
namespace RTC
{
  /*!
   * @if jp
   * @class RingBuffer
   * @brief Oobt@NX
   * 
   * w肵̃Oobt@obt@NXB
   * obt@ŜɃf[^i[ꂽꍇAȍ~̃f[^͌Âf[^
   * ㏑B
   * ]āAobt@ɂ͒߂̃obt@̃f[^̂ݕێB
   *
   * )݂̎ł́AԍŌɊi[f[^̂݃obt@ǂݏo\
   *
   * @param DataType obt@Ɋi[f[^^
   *
   * @since 0.4.0
   *
   * @else
   * @class RingBuffer
   * @brief Ring buffer implementation class
   * 
   * This is the buffer implementation class with ring shaped buffer of
   * specified length.
   * If data is stored in the entire buffer, data from now on will be
   * overwritten from old data one by one.
   * Therefore, only the length of latest data is stored in the buffer.
   *
   * Note: In the current implementation, only last stored data can be read
   *       from the buffer.
   *
   * @param DataType Data type to store in the buffer
   *
   * @since 0.4.0
   *
   * @endif
   */
  template <class DataType>
  class RingBuffer
    : public BufferBase<DataType>
  {
  public:
    BUFFERSTATUS_ENUM
    typedef coil::Guard<coil::Mutex> Guard;
    /*!
     * @if jp
     *
     * @brief RXgN^
     * 
     * RXgN^
     * w肳ꂽobt@Ńobt@B
     *
     * @param length obt@
     * 
     * @else
     *
     * @brief Constructor
     * 
     * Constructor.
     * Initialize the buffer by specified buffer length.
     * However, if the specified length is less than two, the buffer should
     * be initialized by two in length.
     *
     * @param length Buffer length
     * 
     * @endif
     */
    RingBuffer(long int length = RINGBUFFER_DEFAULT_LENGTH)
      : m_overwrite(true), m_readback(true),
        m_timedwrite(false), m_timedread(false),
        m_wtimeout(1, 0), m_rtimeout(1, 0),
        m_length(length),
        m_wpos(0), m_rpos(0), m_fillcount(0), m_wcount(0),
        m_buffer(m_length)
    {
      this->reset();
    }
    
    /*!
     * @if jp
     *
     * @brief zfXgN^
     * 
     * zfXgN^B
     * 
     * @else
     *
     * @brief Virtual destractor
     * 
     * Virtual destractor
     * 
     * @endif
     */
    virtual ~RingBuffer(void)
    {
    }
    
    /*!
     * @if jp
     * @brief obt@̐ݒ
     *
     * coil::Properties ŗ^vpeBɂA
     * obt@̐ݒB
     * gpłIvVƈӖ͈ȉ̒ʂ
     *
     * - buffer.length:
     *     obt@̒BRȊO̐lw肳ĂB
     *     łɃobt@gpԂłAĐݒ肳ꂽ̂AׂĂ
     *     |C^B
     *
     * - buffer.write.full_policy:
     *     ㏑邩ǂ̃|V[B
     *     overwrite (㏑), do_nothing (Ȃ), block (ubN)
     *     block w肵ꍇA timeout lw肷΁Aw莞Ԍ
     *     ݕs\ł΃^CAEgB
     *     ftHg  overwrite (㏑)B
     *
     * - buffer.write.timeout:
     *     ^CAEgԂ [sec] Ŏw肷BftHg 1.0 [sec]B
     *     1 sec -> 1.0, 1 ms -> 0.001, ^CAEgȂ -> 0.0
     *
     * - buffer.read.empty_policy:
     *     obt@̂Ƃ̓ǂݏo|V[B
     *     readback (Ō̗vf), do_nothing (Ȃ), block (ubN)
     *     block w肵ꍇA timeout lw肷΁Aw莞Ԍ
     *     ǂݏos\ł΃^CAEgB
     *     ftHg readback (Ō̗vf)B
     *
     * - buffer.read.timeout:
     *     ^CAEg [sec] Ŏw肷BftHg 1.0 [sec]B
     *     1sec -> 1.0, 1ms -> 0.001, ^CAEgȂ -> 0.0
     *
     * @else
     *
     * @endif
     */
    virtual void init(const coil::Properties& prop)
    {
      initLength(prop);
      initWritePolicy(prop);
      initReadPolicy(prop);
    }
    
    /*!
     * @if jp
     *
     * @brief obt@擾
     * 
     * obt@擾B
     * 
     * @return obt@
     * 
     * @else
     *
     * @brief Get the buffer length
     * 
     * Get the buffer length.
     * 
     * @return Buffer length
     * 
     *
     * @endif
     */
    virtual size_t length(void) const
    {
      Guard guard(m_posmutex);
      return m_length;
    }
    
    /*!
     * @if jp
     *
     * @brief obt@̒Zbg
     * 
     * obt@ݒ肷BݒsȏꍇNOT_SUPPORTEDԂB
     * ̎ł BUFFER_OK ԂȂB
     * 
     * @return BUFFER_OK: I
     *         NOT_SUPPORTED: obt@ύXs
     *         BUFFER_ERROR: ُI
     * 
     * @else
     *
     * @brief Get the buffer length
     *
     * Pure virtual function to get the buffer length.
     *
     * @return buffer length
     * 
     * @endif
     */
    virtual ReturnCode length(size_t n)
    {
      m_buffer.resize(n);
      m_length = n;
      this->reset();
      return ::RTC::BufferStatus::BUFFER_OK; //BUFFER_OK;
    }
    
    /*!
     * @if jp
     *
     * @brief obt@̏ԂZbg
     * 
     * obt@̓ǂݏo|C^Ə݃|C^̈ʒuZbgB
     * ̎ł BUFFER_OK ԂȂB
     * 
     * @return BUFFER_OK: I
     *         NOT_SUPPORTED: Zbgs\
     *         BUFFER_ERROR: ُI
     * 
     * @else
     *
     * @brief Get the buffer length
     *
     * Pure virtual function to get the buffer length.
     *
     * @return buffer length
     * 
     * @endif
     */ 
    virtual ReturnCode reset()
    {
      Guard guard(m_posmutex);
      m_fillcount = 0;
      m_wcount = 0;
      m_wpos = 0;
      m_rpos = 0;
      return ::RTC::BufferStatus::BUFFER_OK;
    }
    
    
    
    //----------------------------------------------------------------------
    /*!
     * @if jp
     *
     * @brief obt@݂̌̏ݗvf̃|C^
     * 
     * obt@݂̌̏ݗvf̃|C^܂́An̃|C^Ԃ
     * 
     * @param  n ݃|C^ + n ̈ʒũ|C^ 
     * @return ݈ʒũ|C^
     * 
     * @else
     *
     * @brief Get the buffer length
     *
     * Pure virtual function to get the buffer length.
     *
     * @return buffer length
     * 
     * @endif
     */ 
    virtual DataType* wptr(long int n = 0) 
    {
      Guard guard(m_posmutex);
      return &m_buffer[(m_wpos + n + m_length) % m_length];
    }
    
    /*!
     * @if jp
     *
     * @brief ݃|C^i߂
     * 
     * ݂݈̏ʒũ|C^ n i߂B
     * ݉\ȗvfȏ̐lw肵ꍇAPRECONDITION_NOT_MET
     * ԂB
     * 
     * @param  n ݃|C^ + n ̈ʒũ|C^ 
     * @return BUFFER_OK:            I
     *         PRECONDITION_NOT_MET: n > writable()
     * 
     * @else
     *
     * @brief Get the buffer length
     *
     * Pure virtual function to get the buffer length.
     *
     * @return buffer length
     * 
     * @endif
     */ 
    virtual ReturnCode advanceWptr(long int n = 1)
    {
      // n > 0 :
      //     n satisfies n <= writable elements
      //                 n <= m_length - m_fillcout
      // n < 0 : -n = n'
      //     n satisfies n'<= readable elements
      //                 n'<= m_fillcount
      //                 n >= - m_fillcount
      Guard guard(m_posmutex);
      if ((n > 0 && n > static_cast<long int>(m_length - m_fillcount)) ||
          (n < 0 && n < static_cast<long int>(-m_fillcount)))
        {
          return ::RTC::BufferStatus::PRECONDITION_NOT_MET;
        }

      m_wpos = (m_wpos + n + m_length) % m_length;
      m_fillcount += n;
      m_wcount += n;
      return ::RTC::BufferStatus::BUFFER_OK;
    }
    /*!
     * @if jp
     *
     * @brief obt@Ƀf[^
     * 
     * obt@Ƀf[^ށB݃|C^̈ʒu͕ύXȂB
     * ̎ł͏ BUFFER_OK ԂB
     * 
     * @param value ݑΏۃf[^
     *
     * @return BUFFER_OK: I
     *         BUFFER_ERROR: ُI
     * 
     * @else
     *
     * @brief Write data into the buffer
     *
     * Pure virtual function to write data into the buffer.
     * Always BUFFER_OK will be returned in this implementation.
     *
     * @param value Target data to write.
     *
     * @return BUFFER_OK:    Successful
     *         BUFFER_ERROR: Failed
     *
     * @endif
     */
    virtual ReturnCode put(const DataType& value)
    {
      Guard guard(m_posmutex);
      m_buffer[m_wpos] = value;
      return ::RTC::BufferStatus::BUFFER_OK;
    }
    
    /*!
     * @if jp
     *
     * @brief obt@ɏ
     * 
     * ŗ^ꂽf[^obt@ɏށB
     *
     * 2(sec)A3(nsec)w肳ĂȂꍇAobt@t
     * ̏݃[h (overwrite, do_nothing, block)  init() Őݒ
     * ꂽ[hɏ]B
     *
     * 2(sec) Ɉw肳ꂽꍇ́Ainit() Őݒ肳ꂽ[h
     * Ɋւ炸Ablock [hƂȂAobt@tԂłΎw莞
     * Ԃ܂A^CAEgB3(nsec)͎w肳Ȃꍇ0Ƃ
     * B^CAEg҂ɁAǂݏoXbhŃobt@
     * ǂݏo΁AubLO͉f[^܂B
     *
     * ݎɃobt@(empty)ԂŁAʂ̃Xbhblock[h
     * œǂݏo҂ĂꍇAsignal𔭍sēǂݏõub
     * LOB
     * 
     * @param value ݑΏۃf[^
     * @param sec   ^CAEg sec  (default -1: )
     * @param nsec  ^CAEg nsec (default 0)
     * @return BUFFER_OK            I
     *         BUFFER_FULL          obt@t
     *         TIMEOUT              ݂^CAEg
     *         PRECONDITION_NOT_MET ݒُ
     * 
     * @else
     *
     * @brief Write data into the buffer
     * 
     * Write data which is given argument into the buffer.
     *
     * @param value Target data for writing
     *
     * @return Writing result (Always true: writing success is returned)
     * 
     * @endif
     */
    virtual ReturnCode write(const DataType& value,
                             long int sec = -1, long int nsec = 0)
    {
      {
      Guard guard(m_full.mutex);
        
      if (full())
        {
          
          bool timedwrite(m_timedwrite);
          bool overwrite(m_overwrite);

          if (!(sec < 0)) // if second arg is set -> block mode
            {
              timedwrite = true;
              overwrite  = false;
            }

          if (overwrite && !timedwrite)       // "overwrite" mode
            {
              advanceRptr();
            }
          else if (!overwrite && !timedwrite) // "do_nothing" mode
            {
              return ::RTC::BufferStatus::BUFFER_FULL;
            }
          else if (!overwrite && timedwrite)  // "block" mode
            {
              if (sec < 0)
                {
                  sec = m_wtimeout.sec();
                  nsec = m_wtimeout.usec() * 1000;
                }
              //  true: signaled, false: timeout
              if (!m_full.cond.wait(sec, nsec))
                {
                  return ::RTC::BufferStatus::TIMEOUT;
                }
            }
          else                                    // unknown condition
            {
              return ::RTC::BufferStatus::PRECONDITION_NOT_MET;
            }
        }
      }      
    
      put(value);

	  {
		Guard eguard(m_empty.mutex);
		if (empty())
		  {
			// Guard eguard(m_empty.mutex);
			advanceWptr(1);
			m_empty.cond.signal();
		  }
		else
		  {
			advanceWptr(1);
		  }
	  }
      return ::RTC::BufferStatus::BUFFER_OK;
    }
    
    /*!
     * @if jp
     *
     * @brief obt@ɏ݉\ȗvf
     * 
     * obt@ɏ݉\ȗvfԂB
     * 
     * @return ݉\ȗvf
     * 
     * @else
     *
     * @brief Write data into the buffer
     *
     * Pure virtual function to write data into the buffer.
     *
     * @param value Target data to write.
     *
     * @return Result of having written in data (true:Successful, false:Failed)
     *
     * @endif
     */
    virtual size_t writable() const
    {
      Guard guard(m_posmutex);
      return m_length - m_fillcount;
    }
    
    /*!
     * @if jp
     *
     * @brief obt@full`FbN
     * 
     * obt@full`FbNpz֐
     *
     * @return full`FbN(true:obt@fullCfalse:obt@󂫂)
     * 
     * @else
     *
     * @brief Check on whether the buffer is full.
     *
     * Pure virtual function to check on whether the buffer is full.
     *
     * @return True if the buffer is full, else false.
     *
     * @endif
     */
    virtual bool full(void) const
    {
      Guard guard(m_posmutex);
      return m_length == m_fillcount;
    }
    
    //----------------------------------------------------------------------
    /*!
     * @if jp
     *
     * @brief obt@݂̌̓ǂݏovf̃|C^
     * 
     * obt@݂̌̓ǂݏovf̃|C^܂́An̃|C^Ԃ
     * 
     * @param  n ǂݏo|C^ + n ̈ʒũ|C^ 
     * @return ǂݏoʒũ|C^
     * 
     * @else
     *
     * @brief Get the buffer length
     *
     * Pure virtual function to get the buffer length.
     *
     * @return buffer length
     * 
     * @endif
     */ 
    virtual DataType* rptr(long int n = 0)
    {
      Guard guard(m_posmutex);
      return &(m_buffer[(m_rpos + n + m_length) % m_length]);
    }
    
    /*!
     * @if jp
     *
     * @brief ǂݏo|C^i߂
     * 
     * ݂̓ǂݏoʒũ|C^ n i߂B
     * 
     * @param  n ǂݏo|C^ + n ̈ʒũ|C^ 
     * @return BUFFER_OK: I
     *         BUFFER_ERROR: ُI
     * 
     * @else
     *
     * @brief Get the buffer length
     *
     * Pure virtual function to get the buffer length.
     *
     * @return buffer length
     * 
     * @endif
     */ 
    virtual ReturnCode advanceRptr(long int n = 1)
    {
      // n > 0 :
      //     n satisfies n <= readable elements
      //                 n <= m_fillcout 
      // n < 0 : -n = n'
      //     n satisfies n'<= m_length - m_fillcount
      //                 n >= m_fillcount - m_length
      Guard guard(m_posmutex);
      if ((n > 0 && n > static_cast<long int>(m_fillcount)) ||
          (n < 0 && n < static_cast<long int>(m_fillcount - m_length)))
        {
          return ::RTC::BufferStatus::PRECONDITION_NOT_MET;
        }

      m_rpos = (m_rpos + n + m_length) % m_length;
      m_fillcount -= n;
      return ::RTC::BufferStatus::BUFFER_OK;
    }
    
    /*!
     * @if jp
     *
     * @brief obt@f[^ǂݏo
     * 
     * obt@f[^ǂ݂Bǂݏo|C^̈ʒu͕ύXȂB
     * 
     * @param value ǂݏof[^
     *
     * @return BUFFER_OK: I
     *         BUFFER_ERROR: ُI
     * 
     * @else
     *
     * @brief Write data into the buffer
     *
     * Pure virtual function to write data into the buffer.
     *
     * @param value Target data to write.
     *
     * @return Result of having written in data (true:Successful, false:Failed)
     *
     * @endif
     */
    virtual ReturnCode get(DataType& value)
    {
      Guard gaurd(m_posmutex);
      value = m_buffer[m_rpos];
      return ::RTC::BufferStatus::BUFFER_OK;
    }
    
    
    /*!
     * @if jp
     *
     * @brief obt@f[^ǂݏo
     * 
     * obt@f[^ǂ݂Bǂݏo|C^̈ʒu͕ύXȂB
     *
     * @return ǂݏof[^
     * 
     * @else
     *
     * @brief Reading data from the buffer
     *
     * @return Read data
     *
     * @endif
     */
    virtual DataType& get()
    {
      Guard gaurd(m_posmutex);
      return m_buffer[m_rpos];
    }
    
    
    /*!
     * @if jp
     *
     * @brief obt@ǂݏo
     * 
     * obt@Ɋi[ꂽf[^ǂݏoB
     *
     * 2(sec)A3(nsec)w肳ĂȂꍇAobt@
     * Ԃł̓ǂݏo[h (readback, do_nothing, block)  init() Ő
     * 肳ꂽ[hɏ]B
     *
     * 2(sec) Ɉw肳ꂽꍇ́Ainit() Őݒ肳ꂽ[h
     * Ɋւ炸Ablock [hƂȂAobt@ԂłΎw莞
     * ҂A^CAEgB3(nsec)͎w肳Ȃꍇ0ƂĈ
     * B^CAEg҂ɁA݃XbhŃobt@֏
     * ΁AubLO͉f[^ǂ݂B
     *
     * ǂݏoɃobt@(empty)ԂŁAʂ̃Xbhblock[h
     * ŏݑ҂ĂꍇAsignal𔭍sďݑ̃ubL
     * OB
     * 
     * @param value ǂݏoΏۃf[^
     * @param sec   ^CAEg sec  (default -1: )
     * @param nsec  ^CAEg nsec (default 0)
     * @return BUFFER_OK            I
     *         BUFFER_EMPTY         obt@
     *         TIMEOUT              ݂^CAEg
     *         PRECONDITION_NOT_MET ݒُ
     * 
     * @else
     *
     * @brief Readout data from the buffer
     * 
     * Readout data stored into the buffer.
     * 
     * @param value Readout data
     *
     * @return Readout result (Always true: readout success is returned)
     * 
     * @endif
     */
    virtual ReturnCode read(DataType& value,
                            long int sec = -1, long int nsec = 0)
    {
      {
      Guard gaurd(m_empty.mutex);
      
      if (empty())
        {
          bool timedread(m_timedread);
          bool readback(m_readback);

          if (!(sec < 0)) // if second arg is set -> block mode
            {
              timedread = true;
              readback  = false;
              sec = m_rtimeout.sec();
              nsec = m_rtimeout.usec() * 1000;
            }

          if (readback && !timedread)       // "readback" mode
            {
              if (!(m_wcount > 0))
                {
                  return ::RTC::BufferStatus::BUFFER_EMPTY;
                }
              advanceRptr(-1);
            }
          else if (!readback && !timedread) // "do_nothing" mode
            {
              return ::RTC::BufferStatus::BUFFER_EMPTY;
            }
          else if (!readback && timedread)  // "block" mode
            {
              if (sec < 0)
                {
                  sec = m_rtimeout.sec();
                  nsec = m_rtimeout.usec() * 1000;
                }
              //  true: signaled, false: timeout
              if (!m_empty.cond.wait(sec, nsec))
                {
                  return ::RTC::BufferStatus::TIMEOUT;
                }
            }
          else                                    // unknown condition
            {
              return ::RTC::BufferStatus::PRECONDITION_NOT_MET;
            }
        }
      }
      
      get(value);

	  {
		Guard fguard(m_full.mutex);
		if (full())
		  {
			// Guard fguard(m_full.mutex);
			advanceRptr(1);
			m_full.cond.signal();
		  }
		else
		  {
			advanceRptr(1);
		  }
	  }
      return ::RTC::BufferStatus::BUFFER_OK;
    }
    
    /*!
     * @if jp
     *
     * @brief obt@ǂݏo\ȗvf
     * 
     * obt@ǂݏo\ȗvfԂB
     * 
     * @return ǂݏo\ȗvf
     *
     * @return BUFFER_OK: I
     *         BUFFER_ERROR: ُI
     * 
     * @else
     *
     * @brief Write data into the buffer
     *
     * Pure virtual function to write data into the buffer.
     *
     * @param value Target data to write.
     *
     * @return Result of having written in data (true:Successful, false:Failed)
     *
     * @endif
     */
    virtual size_t readable() const
    {
      Guard guard(m_posmutex);
      return m_fillcount;
    }
    
    /*!
     * @if jp
     *
     * @brief obt@empty`FbN
     * 
     * obt@empty`FbNpz֐
     *
     * @return empty`FbN(true:obt@emptyCfalse:obt@f[^)
     * 
     * @else
     *
     * @brief Check on whether the buffer is empty.
     *
     * Pure virtual function to check on whether the buffer is empty.
     *
     * @return True if the buffer is empty, else false.
     *
     * @endif
     */
    virtual bool empty(void) const
    {
      Guard guard(m_posmutex);
      return m_fillcount == 0;
    }
    
  private:
    inline void initLength(const coil::Properties& prop)
    {
      if (!prop["length"].empty())
        {
          size_t n;
          if (coil::stringTo(n, prop["length"].c_str()))
            {
              if (n > 0)
                {
                  this->length(n);
                }
            }
        }
    }
    
    inline void initWritePolicy(const coil::Properties& prop)
    {
      std::string policy(prop["write.full_policy"]);
      coil::normalize(policy);
      if (policy == "overwrite")
        {
          m_overwrite = true;
          m_timedwrite = false;
        }
      else if (policy == "do_nothing")
        {
          m_overwrite = false;
          m_timedwrite = false;
        }
      else if (policy == "block")
        {
          m_overwrite = false;
          m_timedwrite = true;
          
          double tm;
          if (coil::stringTo(tm, prop["write.timeout"].c_str()))
            {
              if (!(tm < 0))
                {
                  m_wtimeout = tm;
                }
            }
        }
    }
    
    inline void initReadPolicy(const coil::Properties& prop)
    {
      std::string policy(prop["read.empty_policy"]);
      if (policy == "readback")
        {
          m_readback = true;
          m_timedread = false;
        }
      else if (policy == "do_nothing")
        {
          m_readback = false;
          m_timedread = false;
        }
      else if (policy == "block")
        {
          m_readback = false;
          m_timedread = true;
          double tm;
          if (coil::stringTo(tm, prop["read.timeout"].c_str()))
            {
              m_rtimeout = tm;
            }
        }
    }
    
  private:
    /*!
     * @if jp
     * @brief ㏑tO
     * @else
     * @brief Overwrite flag
     * @endif
     */
    bool m_overwrite;

    /*!
     * @if jp
     * @brief ǂݖ߂tO
     * @else
     * @brief Readback flag
     * @endif
     */
    bool m_readback;

    /*!
     * @if jp
     * @brief ^CAEgt݃tO
     * @else
     * @brief Timedwrite flag
     * @endif
     */
    bool m_timedwrite;
    /*!
     * @if jp
     * @brief ^CAEgtǂݏotO
     * @else
     * @brief Timedread flag
     * @endif
     */
    bool m_timedread;

    /*!
     * @if jp
     * @brief ݎ^CAEg
     * @else
     * @brief Timeout time for writing
     * @endif
     */
    coil::TimeValue m_wtimeout;

    /*!
     * @if jp
     * @brief ǂݏo^CAEg
     * @else
     * @brief Timeout time of reading
     * @endif
     */
    coil::TimeValue m_rtimeout;

    /*!
     * @if jp
     * @brief obt@
     * @else
     * @brief Buffer length
     * @endif
     */
    size_t m_length;

    /*!
     * @if jp
     * @brief ݃|C^
     * @else
     * @brief pointer to write
     * @endif
     */
    size_t m_wpos;

    /*!
     * @if jp
     * @brief ǂݏo|C^
     * @else
     * @brief poitner to read
     * @endif
     */
    size_t m_rpos;

    /*!
     * @if jp
     * @brief FillJEg
     * @else
     * @brief Fill count
     * @endif
     */
    size_t m_fillcount;

    /*!
     * @if jp
     * @brief ݃JEg
     * @else
     * @brief Counter for writing
     * @endif
     */
    size_t m_wcount;

    /*!
     * @if jp
     * @brief obt@z
     * @else
     * @brief baffer array
     * @endif
     */
    std::vector<DataType> m_buffer;
    
    /*!
     * @if jp
     * @brief ϐ\
     * @else
     * @brief struct for condition variable
     * @endif
     */
    struct condition
    {
      condition() : cond(mutex) {}
      coil::Condition<coil::Mutex> cond;
      coil::Mutex mutex;
    };
    
    /*!
     * @if jp
     * @brief ʒuϐ~[ebNX
     * @else
     * @brief mutex for position variable
     * @endif
     */
    mutable coil::Mutex m_posmutex;

    /*!
     * @if jp
     * @brief ϐ
     * @else
     * @brief empty condition variable
     * @endif
     */
    condition m_empty;

    /*!
     * @if jp
     * @brief tϐ
     * @else
     * @brief full condition variable
     * @endif
     */
    condition m_full;
  };
}; // namespace RTC

#endif // RTC_RINGBUFFER_H
